(* File: state.ml *)
(* Author: Chad E Brown *)
(* Created: October 2010 *)

open String
open Syntax
open Mizartypeinfo
open Log
open Error

let currentelaboratename = ref "conj"

(*** FOF BEGIN ***)
let fofconjcnt = ref 0
    
let plit_fo_str a =
  if a > 0 then
    (string_of_int a)
  else
    "~ " ^ (string_of_int (- a))

let rec printf_fotrm s m =
  match m with
  | FOVar(x) -> Printf.fprintf s "%s" x
  | FOFun(f,[]) -> Printf.fprintf s "%s" f
  | FOFun(f,(a::al)) ->
      begin
	Printf.fprintf s "%s(" f;
	printf_fotrm s a;
	List.iter (fun t -> Printf.fprintf s ","; printf_fotrm s t) al;
	Printf.fprintf s ")";
      end

let rec printf_fof s m =
  match m with
  | FOEq(t1,t2) ->
      Printf.fprintf s "(";
      printf_fotrm s t1;
      Printf.fprintf s " = ";
      printf_fotrm s t2;
      Printf.fprintf s ")";
  | FOPred(p,[]) ->
      Printf.fprintf s "%s" p;
  | FOPred(p,(a::al)) ->
      Printf.fprintf s "%s(" p;
      printf_fotrm s a;
      List.iter (fun t -> Printf.fprintf s ","; printf_fotrm s t) al;
      Printf.fprintf s ")";
  | FOImp(m1,FOFal) -> (*** handle this as negation because => $false seems to confuse E ***)
      Printf.fprintf s "(~ ";
      printf_fof s m1;      
      Printf.fprintf s ")";
  | FOImp(m1,m2) ->
      Printf.fprintf s "(";
      printf_fof s m1;
      Printf.fprintf s " => ";
      printf_fof s m2;
      Printf.fprintf s ")";
  | FOEqu(m1,m2) ->
      Printf.fprintf s "(";
      printf_fof s m1;
      Printf.fprintf s " <=> ";
      printf_fof s m2;
      Printf.fprintf s ")";
  | FOAll(x,m1) ->
      Printf.fprintf s "( ! [%s] : " x;
      printf_fof s m1;
      Printf.fprintf s ")";
  | FOFal ->
      Printf.fprintf s "$false"

let rec printf_fof_neg_alls s m =
  match m with
  | FOAll(x,m1) ->
      Printf.fprintf s "( ? [%s] : " x;
      printf_fof_neg_alls s m1;
      Printf.fprintf s ")"
  | _ ->
      Printf.fprintf s "(~ ";
      printf_fof s m;
      Printf.fprintf s ")"
(*** FOF END ***)  

let timeout : float option ref = ref None
let hardtimeout : float option ref = ref None
let problemfile = ref ""

type probitem =
  | ProbDef of string * stp * trm * (string * string) list * float
  | ProbAx of string * string * trm * (string * string) list * float
  | ProbConj of string * trm * (string * string) list * float
let probsig : probitem list ref = ref []

let conjecturename : string ref = ref "claim"
let conjecture : (trm * trm) option ref = ref None
type proofkind = TSTP | CoqScript | CoqSPfTerm | HOCore | Model | ModelTrue | IsarScript | PfInfo | PfUseful | PfFormdeps
let mkproofterm = ref None
let mkprooftermp () = false
let pfusefulout = ref None
let pfformdepsout = ref None
let slave = ref false
let name_base : (string,unit) Hashtbl.t = Hashtbl.create 10
let name_base_list : string list ref = ref []
let name_tp : (string,stp) Hashtbl.t = Hashtbl.create 100
let name_trm : (string,(trm * stp) * bool ref) Hashtbl.t = Hashtbl.create 100
let name_trm_list : (string * trm * stp) list ref = ref []
let translucent_defns : bool ref = ref false
let name_def_empty : (string,trm) Hashtbl.t = Hashtbl.create 1
let name_def : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_def_all : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_def_prenorm : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_hyp : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_hyp_inv : (trm,string * trm) Hashtbl.t = Hashtbl.create 100 (*** associate a normalized/logic normalized term with the name of the assumption it came from and its original prenormalized form ***)
let assumption_lit : (int,trm * trm) Hashtbl.t = Hashtbl.create 100 (*** associate assumption literals with their term after preprocessing and before preprocessing ***)

(** Replaces all occurences of 'Neg' by 'implies False' **)
let rec negnorm1 m =
  match m with
  | Ap(Neg,m1) ->
      let (n1,_) = negnorm1 m1 in
      (imp n1 False,true)
  | Neg -> (Lam(Prop,imp (DB(0,Prop)) False),true)
  | Ap(m1,m2) ->
      let (n1,b1) = negnorm1 m1 in
      let (n2,b2) = negnorm1 m2 in
      if (b1 || b2) then
	(Ap(n1,n2),true)
      else
	(m,false)
  | Lam(a1,m1) ->
      let (n1,b1) = negnorm1 m1 in
      if b1 then
	(Lam(a1,n1),true)
      else
	(m,false)
  | _ -> (m,false)
(** applies neg- and betanormalization**)
let onlynegnorm m =
  let (n,_) = negnorm1 m in onlybetanorm n
(** applies neg-, beta- and delta-normalization**)
(** partially normalizes, without expanding certain defns (e.g., is_of) **)
let partialnormalize pt = norm name_def (logicnorm pt)
(** applies full satallax normalization**)
let normalize pt = norm name_def_all (logicnorm pt)
let belnorm pt = norm name_def_empty (logicnorm pt)

(*** March 31, 2011 - Chad - THF policy regarding definitions. (See comments before declare_definition_real below. ***)
exception TreatAsAssumption

let next_fresh_name : int ref = ref 0

let rec get_fresh_name a =
  let x = make_fresh_name !next_fresh_name in
  incr next_fresh_name;
  if ((Hashtbl.mem name_base x) || (Hashtbl.mem name_tp x)) then
    get_fresh_name a
  else
    let xa = Name(x,a) in
    begin
      Hashtbl.add name_tp x a;
      Hashtbl.add name_trm x ((xa,a), ref false);
      name_trm_list := (x,xa,a)::!name_trm_list;
      (x,xa)
    end


let initial_branch : trm list ref = ref []
let initial_branch_prenorm : trm list ref = ref []

let processed : (trm,int) Hashtbl.t = Hashtbl.create 100

let patoms : (string,int * (trm list)) Hashtbl.t = Hashtbl.create 10
let natoms : (string,int * (trm list)) Hashtbl.t = Hashtbl.create 10

let pchoiceatoms : (stp,int * (trm list)) Hashtbl.t = Hashtbl.create 10
let nchoiceatoms : (stp,int * (trm list)) Hashtbl.t = Hashtbl.create 10

let peqns : (string,int * trm * trm) Hashtbl.t = Hashtbl.create 10
let neqns : (string,int * trm * trm) Hashtbl.t = Hashtbl.create 10

let univpreds : (stp,(int * trm)) Hashtbl.t = Hashtbl.create 10
let instantiations : (stp,(trm,unit) Hashtbl.t) Hashtbl.t = Hashtbl.create 10
let instantiationslist : (stp,trm list) Hashtbl.t = Hashtbl.create 10
let default_elts : (string,trm) Hashtbl.t = Hashtbl.create 10

let set_default_elt aname x = Hashtbl.add default_elts aname x

let set_default_elt_if_needed a x =
  match a with
  | Base(aname) ->
      if (not (Hashtbl.mem default_elts aname)) then set_default_elt aname x
  | _ -> ()

let default_elt aname =
  try
    Hashtbl.find default_elts aname
  with
  | Not_found ->
      let a = Base aname in
      let (_,x) = get_fresh_name a in
      begin
	set_default_elt aname x;
	x
      end

let default_elt_p aname =
  Hashtbl.mem default_elts aname

let get_instantiations a =
  try
    Hashtbl.find instantiationslist a
  with Not_found -> []

let known_instantiation a m =
  try
    let h = Hashtbl.find instantiations a in
    Hashtbl.find h m;
    true
  with Not_found -> false

let filtered : (int,unit) Hashtbl.t = Hashtbl.create 10

let part_of_conjecture : (trm,unit) Hashtbl.t = Hashtbl.create 10

type namecategory =
    ChoiceOp of int * int * stp list * trm list (*** (i,n,sigmal,tl) where length of sigmal and tl are n, 0 <= i < n, Name has type (sigmal -> o) -> sigmal[i], and for each j:{0,...,n-1} tl[j] is the jth component of the n-ary choice operator (in particular, tl[i] is this one) ***)
   | DescrOp of int * int * stp list * trm list (*** (i,n,sigmal,tl) where length of sigmal and tl are n, 0 <= i < n, Name has type (sigmal -> o) -> sigmal[i], and for each j:{0,...,n-1} tl[j] is the jth component of the n-ary description operator (in particular, tl[i] is this one) ***)
   | IfThenElse of int * int * stp list (*** (i,n,sigmal) where length of sigmal is n, 0 <= i < n, Name has type o -> sigmal -> sigmal -> sigmal[i] ***)
   | ReflexiveBinary
   | IrreflexiveBinary
   | SymmetricBinary
   | ReflexiveSymmetricBinary
   | IrreflexiveSymmetricBinary

let constrainedName : (string,namecategory) Hashtbl.t = Hashtbl.create 10

let decomposable x =
   try
   let c = Hashtbl.find constrainedName x in (*** Some categorized names are decomposable and some are not ***)
   match c with
   | IfThenElse _ -> false (*** TO DO: Check that n-ary if-then-else need not be decomposable ***)
   (*** TO DO: Should *Binary cases be decomposable? ***)
   | _ -> true
   with Not_found -> true (*** A name is decomposable by default ***)

(*** Set completep to false if I use a mode that makes the search incomplete, so that failure does not imply satisfiability.
     Currently there is no such mode. - Chad, Oct 2010 ***)
let completep = ref true

let get_timeout_default x = match (!timeout) with Some y -> y | None -> x

let st_include_fun : (string -> unit) ref = ref (fun x -> raise (GenericError("Bug when including file " ^ x)))
let st_find_read_thf_fun : (string -> string -> unit) ref = ref (fun d x -> raise (GenericError("Bug when including file " ^ x)))

let declare_base_type a =
  Hashtbl.add name_base a ();
  name_base_list := (a::!name_base_list)

let st_to_stp m =
  begin
    try
      to_stp m
    with
    | DeclareInd ->
	begin (*** Declare the base type i the first time it's seen.  Use a different name if an 'i' has already been used. ***)
	  declare_base_type "$i";
	  to_stp m
	end
  end

let st_to_trm_given_stp m tp =
  begin
    try
      let (n,_) = to_trm name_trm [] m (Some tp) in n
    with
    | DeclareInd ->
	begin (*** Declare the base type i the first time it's seen.  Use a different name if an 'i' has already been used. ***)
	  declare_base_type "$i";
	  let (n,_) = to_trm name_trm [] m (Some tp) in n
	end
  end

let st_to_trm m =
  begin
    try
      to_trm name_trm [] m None
    with
    | DeclareInd ->
	begin (*** Declare the base type i the first time it's seen.  Use a different name if an 'i' has already been used. ***)
	  declare_base_type "$i";
	  to_trm name_trm [] m None
	end
  end

let is_of_names : string list ref = ref []
let all_of_names : string list ref = ref []
let is_of_name : (string,unit) Hashtbl.t = Hashtbl.create 1
let all_of_name : (string,unit) Hashtbl.t = Hashtbl.create 1

let is_of_p m =
  match m with
  | Name(x,Ar(Base("$i"),Ar(Ar(Base("$i"),Prop),Prop))) -> Hashtbl.mem is_of_name x
  | Lam(Base("$i"),Lam(Ar(Base("$i"),Prop),Ap(DB(0,Ar(Base("$i"),Prop)),DB(1,Base("$i"))))) -> true
  | _ -> false
	
let all_of_p m =
  match m with
  | Name(x,Ar(Ar(Base("$i"),Prop),Ar(Ar(Base("$i"),Prop),Prop))) -> Hashtbl.mem all_of_name x
  | Lam(Ar(Base("$i"),Prop),Lam(Ar(Base("$i"),Prop),Ap(Forall(Base("$i")),Lam(Ar(Base("$i"),Prop),Ap(Ap(Imp,Ap(Ap(m1,DB(0,Base("$i"))),DB(2,Ar(Base("$i"),Prop)))),DB(1,Ar(Base("$i"),Prop))))))) ->
      is_of_p m1
  | _ -> false

let rec ontology_prop_p m =
  match m with
  | Ap(Ap(Name(x,Ar(Ar(Base("$i"),Prop),Ar(Ar(Base("$i"),Prop),Prop))),m1),m2) when Hashtbl.mem all_of_name x -> ontology_prop_p (gen_lam_body (Base("$i")) m2)
  | Ap(Ap(Name(x,Ar(Base("$i"),Ar(Ar(Base("$i"),Prop),Prop))),m1),m2) when Hashtbl.mem is_of_name x -> true
  | Ap(Forall(a),m1) -> ontology_prop_p (gen_lam_body a m1)
  | Ap(Ap(Imp,m1),m2) -> ontology_prop_p m2
  | _ -> false

let declare_typed_constant (name:string) (role:string) (m:pretrm) (al:(string * string) list) =
  begin
    if (!verbosity > 4) then (Printf.printf "declare_typed_constant %s %s\n%s\n" name role (pretrm_str m); flush stdout);
    if (!verbosity > 20) then (Printf.printf "annotations:\n"; List.iter (fun (a,b) -> Printf.printf "%s: %s\n" a b) al; flush stdout);
    match m with
      POf(PName(x),m) ->
	begin
	  match m with
	    PType -> (*** Actually unused. - April 20 2012 ***)
	      if (Hashtbl.mem name_base x) then raise (Redeclaration x);
	      if (Hashtbl.mem name_tp x) then raise (Redeclaration x);
	      declare_base_type x
	  | PName "$type" -> (*** The case that's used. Added April 20 2012 ***)
	      if (Hashtbl.mem name_base x) then raise (Redeclaration x);
	      if (Hashtbl.mem name_tp x) then raise (Redeclaration x);
	      declare_base_type x
	  | _ ->
	      let tp = st_to_stp m in
	      if (Hashtbl.mem name_base x) then raise (Redeclaration x);
	      if (Hashtbl.mem name_tp x) then raise (Redeclaration x);
	      Hashtbl.add name_tp x tp;
	      Hashtbl.add name_trm x ((Name(x,tp),tp),ref false);
	      name_trm_list := (x,Name(x,tp),tp)::!name_trm_list
	end
    | _ -> raise (GenericError ("Incorrect format for type declaration " ^ name))
  end

let rec tm_stats m =
  match m with
  | Name(x,_) -> Printf.printf "(NAME \"%s\")\n" x
  | Forall(a) -> Printf.printf "(FORALL \"%s\")\n" (stp_str a)
  | Eq(a) -> Printf.printf "(FORALL \"%s\")\n" (stp_str a)
  | Ap(m,n) -> tm_stats m; tm_stats n
  | Lam(a,m) -> tm_stats m
  | _ -> ()

let printfof = ref true
    
let rec init_probitem m =
  match m with
  | ProbDef(x,a,m,al,w) ->
      begin
	try
	    let m0 = logicnorm m in
	    let m2 = norm name_def m0 in
	    if (!verbosity > 20) then Printf.printf "name_def %s %s\n" x (trm_str m2);
	    Hashtbl.add name_def_all x (norm name_def_all m2);
	with TreatAsAssumption ->
	  init_probitem (ProbAx(x,"hypothesis",Ap(Ap(Eq(a),Name(x,a)),m),al,w))
      end
  | ProbAx(name,role,tm,al,w) ->
      begin
	let m0 = logicnorm tm in
	let tmn = norm name_def m0 in
	Hashtbl.add name_hyp name tmn;
	Hashtbl.add name_hyp_inv tmn (name,tm);
	if (!verbosity > 20) then Printf.printf "name_hyp %s %s\n" name (trm_str tmn);
	initial_branch_prenorm := tm::(!initial_branch_prenorm);
	initial_branch := tmn::(!initial_branch)
      end
  | ProbConj(name,tm,al,w) ->
      begin
	conjecturename := name;
	if (name != "claim") then
	  ignore (coqify_name name coq_hyp_names coq_used_names);
	let ntm = Ap(Neg,tm) in
	let ntmn = norm name_def (logicnorm ntm) in
	initial_branch_prenorm := (ntm::(!initial_branch_prenorm));
	if (!verbosity > 20) then Printf.printf "name_conj negated %s %s\n" name (trm_str ntmn);
	initial_branch := (ntmn::(!initial_branch));
	conjecture := Some (tm,ntmn);
	Hashtbl.add part_of_conjecture ntmn ()
      end

let defpreweight m al = 2.0

let axpreweight m al =
  if ontology_prop_p m then 1.0 else 3.0

let conjpreweight m al = 0.0

(***
 THF official policy requires that all names are given a type before use,
 that definitions can be circular (in which case they must be treated as equations),
 and that definitions may appear *after* it's already been used in an assumption.
 In order to comply with this policy *and* do something reasonable in the 'usual' case,
 I do the following:

 1. I keep a hashtable (occurred_names) of typed names that have been used in an assumption.
 2. When I encounter a definition, if it's of the form (x = t), I first parse t.  If x has been used,
    then I include the equation as an axiom.  Otherwise, x is treated as a definition and expanded away.

 NOTE: If I wanted to be strict, I would exit if the definition is given without the name being given a type,
 but I will not require the type to be given.  I will print a warning if the verbosity is > 0.
 ***)
let declare_definition_real x m al =
  if (Hashtbl.mem name_base x) then raise (Redeclaration x);
  if (Hashtbl.mem name_def x) then raise TreatAsAssumption; (*** treat it as an assumption per THF policy. - Mar 31, 2011 - Chad ***)
(*** raise (Redeclaration x); ***)
  try
    let tp = Hashtbl.find name_tp x in
    let tm = belnorm (st_to_trm_given_stp m tp) in
    try 
      let (_,r) = Hashtbl.find name_trm x in
      if (!r) then
	raise TreatAsAssumption (*** It's already been used, treat 'definition' as an assumption. ***)
      else
	raise Not_found
    with
    | Not_found ->
	begin
	  Hashtbl.add name_def_prenorm x tm;
	  if is_of_p tm then
	    (is_of_names := x::!is_of_names; Hashtbl.add is_of_name x ())
	  else if all_of_p tm then
	    (all_of_names := x::!is_of_names; Hashtbl.add all_of_name x ());
	  let w = defpreweight tm al in
	  probsig := ProbDef(x,tp,tm,al,w)::!probsig
	end
  with
  | Not_found ->
      begin (*** Giving a definition without giving it's type isn't allowed in THF anymore.  I'm still allowing it.  ***)
	let (tm,tp) = st_to_trm m in
	let tm = belnorm tm in
	Hashtbl.add name_tp x tp;
	Hashtbl.add name_trm x ((Name(x,tp),tp),ref false);
	name_trm_list := (x,Name(x,tp),tp)::!name_trm_list;
	Hashtbl.add name_def_prenorm x tm;
	let w = defpreweight tm al in
	probsig := ProbDef(x,tp,tm,al,w)::!probsig
      end

let rec prehead m a =
  match m with
  | PAp(m1,_) -> prehead m1 (Ar(Base("$i"),a))
  | PName(x) -> (x,a)
  | _ -> raise Not_found
    
let rec prelhshead m =
  match m with
  | PAll(_,m1) -> prelhshead m1
  | PAp(PAp(PImplies,_),m2) -> prelhshead m2
  | PAp(PAp(PIff,m1),_) -> prehead m1 Prop
  | PAp(PAp(PEq,m1),_) -> prehead m1 (Base("$i"))
  | _ -> raise Not_found

let npnum_p x = String.length x > 4 && String.sub x 0 4 = "np__" && (let c = Char.code x.[4] in c >= 48 && c <= 57)

let rec declare_thf_logic_formula (name:string) (role:string) (m:pretrm) (al:(string * string) list) =
  begin
    if !verbosity > 4 then (Printf.printf "declare_thf_logic_formula %s %s\n" name role; flush stdout);
    if (!verbosity > 20) then (Printf.printf "annotations:\n"; List.iter (fun (a,b) -> Printf.printf "%s: %s\n" a b) al; flush stdout);
    if ((role = "axiom") || (role = "hypothesis") || (role = "assumption") || (role = "lemma") || (role = "theorem") || (role = "corollary")) then
      begin
	if (Hashtbl.mem name_hyp name) then raise (Redeclaration name);
	let tm = st_to_trm_given_stp m Prop in
	let tm = belnorm tm in
	let w = axpreweight tm al in
	if !collecttypingaxioms then
	  begin
	    try
	      let (i,pl,ml) = mizar_type_fact1_p tm 0 [] in
	      if ml = [] then raise Not_found; (*** ignore those which have no conclusion (just 'true') ***)
	      names_in_mizar_type_assumptions := [];
	      mizar_hook_names := [];
	      if mizar_type_fact2o_p tm then
		let hooknames2 = !mizar_hook_names in
		begin
		  names_in_mizar_type_assumptions := [];
		  mizar_hook_names := [];
		  if mizar_type_fact3o_p tm then
		    typingaxiomslist := MizarTypeHierarchyStrict(name,!mizar_hook_names,i,pl,ml)::!typingaxiomslist
		  else
		    typingaxiomslist := MizarTypeHierarchy(name,hooknames2,i,pl,ml)::!typingaxiomslist
		end
	      else
		begin
		  names_in_mizar_type_assumptions := [];
		  mizar_hook_names := [];
		  if mizar_type_fact4o_p tm then
		    typingaxiomslist := MizarFunctionType(name,!mizar_hook_names,i,pl,ml)::!typingaxiomslist
		  else
		    typingaxiomslist := MizarTypingGeneric(name,i,pl,ml)::!typingaxiomslist
		end
	    with Not_found ->
	      try
		names_in_mizar_type_assumptions := [];
		let (i,pl,f,s,def) = mizar_type_fact5o_p tm 0 [] in
		typingaxiomslist := MizarRedefinition(name,i,pl,f,s,def)::!typingaxiomslist
	      with Not_found ->
		Printf.printf "WARNING: %s is not a Mizar typing axiom.\n" name;
	  end
	else
	  probsig := ProbAx(name,role,tm,al,w)::!probsig
      end
    else if (role = "conjecture") then
      begin
	match (!conjecture) with
	| Some _ -> raise (GenericError "Problem file has more than one conjecture.")
	| None ->
	    let tm = st_to_trm_given_stp m Prop in
	    let tm = belnorm tm in
	    let w = conjpreweight tm al in
	    probsig := ProbConj(name,tm,al,w)::!probsig
      end
    else if (role = "elaborateset" || role = "elaborateprop"  || role = "elaborate") then
      begin (*** given a proposition to elaborate ***)
	begin
	  try
	    ignore (to_stp PIndTp)
	  with DeclareInd -> declare_base_type "$i"
	end;
	try
	  let metab : (pretrm,trm * stp * ((string * (trm list)) list)) Hashtbl.t = Hashtbl.create 100 in
	  let melabh : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mtph : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mtpc : (trm,unit) Hashtbl.t = Hashtbl.create 100 in
	  let mccv : (trm,unit) Hashtbl.t = Hashtbl.create 100 in
	  let mccp : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mccb : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mccf : (trm,trm) Hashtbl.t = Hashtbl.create 100 in
	  let mcc = (mccv,mccp,mccb,mccf) in
	  set_timer !elaboratetimeout;
	  currentelaboratename := name;
	  Hashtbl.clear main_assumed_sorts;
	  Hashtbl.clear main_queried_sorts;
	  assume_needed_sorts := 0;
	  let varh : (string,unit) Hashtbl.t = Hashtbl.create 100 in
	  let varcnt = ref 0 in
	  let mizstate = (metab,melabh,mtph,mtpc,mcc,(varh,ref false,varcnt)) in
	  let (mtm,mtp,frl) =
	    if role = "elaborateset" then
	      let (mtm,frl) = mizar_elaborate_trm_force_stp mizstate [] m (Base("$i")) in
	      (mtm,Base("$i"),frl)
	    else if role = "elaborateprop" then
	      let (mtm,frl) = mizar_elaborate_trm_force_stp mizstate [] m Prop in
	      (mtm,Prop,frl)
	    else
	      mizar_elaborate_trm mizstate [] m
	  in
	  if !printfof && mtp = Prop then
	    begin
	      try
		let (fof,_) = trm_to_foform_stp mtm true in
		incr fofconjcnt;
		Printf.printf "fof(%s,conjecture," !currentelaboratename;
		printf_fof stdout fof;
		Printf.printf ").\n";
	      with NotFO ->
		Printf.printf "#+#++ Could not see %s as fof\n" (trm_str_nice mtm)
	    end;
	  Printf.printf "REMTIME %f\n" (remaining_time());
	  set_timer 0.0;
 (*** return a metatrm with evars for missing arguments, or fail if it's clearly ill-typed ***)
	  Printf.printf "+++ %s elaborated to: %s\n" name (trm_str mtm);
	  let num_assumed = ref 0 in
	  let num_queried = ref 0 in
	  Hashtbl.iter (fun p _ -> Printf.printf "#+#++ Assumed: %s\n" (trm_str p); incr num_assumed) main_assumed_sorts;
	  Hashtbl.iter (fun _ _ -> incr num_queried) main_queried_sorts;
	  let p = if !num_queried > 0 then 1.0 -. (float_of_int !num_assumed /. float_of_int !num_queried) else 1.0 in
	  Printf.printf "#+#++ elaboration success with %d; %d; p = %f\n" !num_assumed !num_queried p;
	with
	| ElaborationSimpleTypeError(mtm,b,a) -> (*** fundamental simple typing problem, definite rejection ***)
	    set_timer 0.0;
	    Printf.printf "--- 5 %s cannot be elaborated: subterm with type %s expected to have type %s\n" name (stp_str b) (stp_str a)
	| ExpectedTypeError(m1,b,a) -> (*** fundamental simple typing problem, definite rejection ***)
	    set_timer 0.0;
	    Printf.printf "--- 5 %s cannot be elaborated: subterm with type %s expected to have type %s\n" name (stp_str b) (stp_str a)
	| GenericSyntaxError(x) -> (*** syntax error, possibly due to insufficient annotations, strong rejection ***)
	    set_timer 0.0;
	    Printf.printf "--- 4 %s cannot be elaborated: Syntax error: %s\n" name x
	| ElaborationFailure(x) -> (*** elaboration failure, which might just be due to the elaboration/type checking algorithm being too weak, possible rejection ***)
	    set_timer 0.0;
	    Printf.printf "--- 1 %s cannot be elaborated: %s\n" name x
	| Timeout ->
	    set_timer 0.0;
	    Printf.printf "--- 3 %s cannot be elaborated: Timeout\n" name
	| e -> (*** other exception, possible rejection ***)
	    set_timer 0.0;
	    Printf.printf "--- 2 %s cannot be elaborated: %s\n" name (Printexc.to_string e)
      end
    else if role = "mizarpattern" then (*** special for mizar functions and relations with missing arguments, not to be treated as logical assumptions ***)
      begin
	try
(*	  Printf.printf "mizarpattern %s\n" name; *)
(*	  Printf.printf "m: %s\n" (pretrm_str m); *)
	  let tm = belnorm (st_to_trm_given_stp m Prop) in
(*	  Printf.printf "tm: %s\n" (trm_str tm); *)
	  let (nname,i,pl,rtp,al,pd,d) = mizar_pattern tm 0 [] in
	  let cl = ref (consts_of_trm [] d) in
	  List.iter (fun n -> cl := consts_of_trm !cl n) pl;
	  List.iter (fun (c,ctp) -> if c.[0] = 'n' && not (npnum_p c) then Printf.printf "WARNING: Surprising name %s found in pattern for %s\n" c nname) !cl;
	  Hashtbl.add mizar_pattern_table nname (i,pl,rtp,al,pd,d);
	  mizar_pattern_list := (nname,i,pl,rtp,al,pd,d)::!mizar_pattern_list;
(*	  Printf.printf "mizar pattern succeeded %s\n" nname *)
	with Not_found ->
	  raise (Failure (name ^ " unexpected mizar pattern"))
      end
    else
      raise (GenericError ("Unknown role " ^ role))
  end
and declare_definition (name:string) (role:string) (m:pretrm) (al:(string * string) list) =
  try
    begin
      if !verbosity > 4 then (Printf.printf "declare_definition %s %s\n" name role; flush stdout);
      if (!verbosity > 20) then (Printf.printf "annotations:\n"; List.iter (fun (a,b) -> Printf.printf "%s: %s\n" a b) al; flush stdout);
      match m with
	PDef(PName(x),m) -> (*** No longer THF syntax. ***)
	  declare_definition_real x m al
      | PAp(PAp(PEq,PName(x)),m) ->
	  declare_definition_real x m al
      | _ -> (*** Treat as an assumption, no matter how it looks. Odd, but OK. This may be too liberal; we haven't decided yet. ***)
	  raise TreatAsAssumption
(*** raise (GenericError ("Incorrect format for definition " ^ name)) ***)
    end
  with
  | TreatAsAssumption ->
      declare_thf_logic_formula name "hypothesis" m al

(*** Code for enumeration of types and terms - Dec 10, 2010 - Chad ***)
let enum_started = ref false
let enum_of_started_ : (stp,unit) Hashtbl.t = Hashtbl.create 5
let enum_of_started a =
  Hashtbl.mem enum_of_started_ a
let enum_of_start a =
  Hashtbl.add enum_of_started_ a ()
let type_continuations_rtp : (stp option,(stp -> int -> unit)) Hashtbl.t = Hashtbl.create 5
let term_continuations_rtp : (stp,(stp list * trm * int -> unit)) Hashtbl.t = Hashtbl.create 5
let usableTypes_rtp : (stp,(stp * int)) Hashtbl.t = Hashtbl.create 5
let usableTypes : (stp * stp * int) list ref = ref []
let usableHeads_rtp : (stp,(stp list * trm * int)) Hashtbl.t = Hashtbl.create 5

let new_type_continuation_rtp ar f =
  Hashtbl.add type_continuations_rtp (Some ar) f

let new_type_continuation f =
  Hashtbl.add type_continuations_rtp None f

let iter_type_continuations_rtp ar a d =
  List.iter (fun f -> f a d) (Hashtbl.find_all type_continuations_rtp (Some ar))

let iter_type_continuations a d =
  List.iter (fun f -> f a d) (Hashtbl.find_all type_continuations_rtp None)

let new_term_continuation_rtp ar f =
  Hashtbl.add term_continuations_rtp ar f

let iter_term_continuations_rtp ar sigmal m p =
  List.iter (fun f -> f (sigmal,m,p)) (Hashtbl.find_all term_continuations_rtp ar)

let new_usable_type_rtp ar a d =
  Hashtbl.add usableTypes_rtp ar (a,d);
  usableTypes := ((ar,a,d)::(!usableTypes))

let usable_types_rtp ar = Hashtbl.find_all usableTypes_rtp ar

let usable_types () = !usableTypes

let new_usable_head_rtp ar sigmal m n = Hashtbl.add usableHeads_rtp ar (sigmal,m,n)

let usable_heads_rtp ar = Hashtbl.find_all usableHeads_rtp ar

